# @file This file tests secureboot_audit.py script
# Command-line tool for inspecting UEFI Secure Boot databases
# Requires "pip install edk2-pytool-extensions"
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Command-line tool for inspecting UEFI Secure Boot databases."""

import argparse
import csv
import ctypes
import json
import logging
import os
import sys
from typing import BinaryIO

import openpyxl
import xlsxwriter

# Some of these libraries are windows only, so we need to import them conditionally
if sys.platform == "win32":
    import pywintypes
    import win32api
    import win32process
    import win32security

# import this from edk2toollib so we can parse the dbx
from edk2toollib.uefi.authenticated_variables_structure_support import (
    EfiSignatureDatabase,
)

logging.basicConfig()
logger = logging.getLogger()
logger.setLevel(logging.INFO)

if sys.platform == "win32":
    KERNEL32 = ctypes.windll.kernel32
    EFI_VAR_MAX_BUFFER_SIZE = 1024 * 1024

DEFAULT_OUTPUT_FOLDER = os.path.join(".", "SecureBootFiles")

# The supported files to retrieve from UEFI and their guids, additional files can be added here
SECUREBOOT_FILES = {
    "dbx": "{d719b2cb-3d3a-4596-a3bc-dad00e67656f}", # EFI_IMAGE_SECURITY_DATABASE_GUID
    "db":  "{d719b2cb-3d3a-4596-a3bc-dad00e67656f}", # EFI_IMAGE_SECURITY_DATABASE_GUID
    "dbt": "{d719b2cb-3d3a-4596-a3bc-dad00e67656f}", # EFI_IMAGE_SECURITY_DATABASE_GUID
    "KEK": "{8BE4DF61-93CA-11d2-AA0D-00E098032B8C}", # EFI_GLOBAL_VARIABLE
    "PK":  "{8BE4DF61-93CA-11d2-AA0D-00E098032B8C}"  # EFI_GLOBAL_VARIABLE
}

KNOWN_CERTIFICATES = [
    "11FB1A56B8CAE09CF7B06AD620A9F13C604D108BDAF443C115FE5848805241E5",  # "Canonical Certificate",
    "F22246C9E65BB4543CB892B6507C062196A788E5A190D2D5A877B94DFDFB2935",  # "Debian Certificate",
]


###################################################################################################
# Helper functions
###################################################################################################


def write_json_file(report: dict, output_file: str) -> None:
    """Writes a JSON file.

    Args:
        report (dict): report to write
        output_file (str): Output file to write to

    Returns:
        None
    """
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    with open(output_file, "w", encoding="utf-8") as f:
        f.write(json.dumps(report, indent=4))

    logger.info("Wrote report to %s", output_file)


def write_xlsx_file(report: dict, output_file: str) -> None:
    """Writes an XLSX file.

    Args:
        report (dict): Data to write
        output_file (str): Output file to write to

    Returns:
        None
    """
    os.makedirs(os.path.dirname(output_file), exist_ok=True)

    workbook = xlsxwriter.Workbook(output_file)

    for page in report.keys():
        worksheet = workbook.add_worksheet(page)

        if "dict" in report[page].keys():
            # Write the header
            col = 0
            row = 0
            headers = ["authenticode_hash"]

            # Build up the field headers in the dictionary
            for auth_hash in report[page]["dict"]:
                for field in report[page]["dict"][auth_hash]:
                    if field not in headers:
                        headers.append(field)
                # Only need to iterate through one hash
                break

            # write the headers
            worksheet.write_row(row, col, headers)
            row += 1

            for auth_hash in report[page]["dict"]:
                # Write the hash
                worksheet.write(row, col, auth_hash)
                col += 1

                # Write the fields
                for field in headers[1:]:  # Skip the hash
                    if field not in report[page]["dict"][auth_hash]:
                        col += 1
                        continue

                    data = report[page]["dict"][auth_hash][field]
                    if isinstance(data, list):
                        data = ", ".join(data)

                    worksheet.write(row, col, data)
                    col += 1

                col = 0
                row += 1

        elif "list" in report[page]:
            row = 0
            headers = ["authenticode_hash"]
            worksheet.write_row(row, col, headers)
            row += 1

            for auth_hash in report[page]["list"]:
                worksheet.write(row, col, auth_hash)
                row += 1

    workbook.close()

    logger.info("Wrote report to %s", output_file)

def convert_row_to_metadata(row: list) -> dict:
    """Converts a row from the csv to a metadata dictionary.

    Args:
        row (list): A row from the csv file

    Returns:
        dict: A dictionary containing the metadata
    """
    convert_arch = {"64-bit": "x86_64", "32-bit": "x86", "64-bit ARM": "arm64"}
    authenticode_hash = row[1].upper()

    def map_cve(cve: str) -> str:
        """Maps a CVE to the correct format.

        Args:
            cve (str): CVE to map

        Returns:
            str: Mapped CVE
        """
        # Intentionally naive mapping of CVE's to the correct format
        if 'CVE' in cve.upper():
            # this condition could have multiple CVE's in it
            return cve
        elif "black lotus" in cve.lower():
            return "CVE-2022-21894"
        else:
            return "N/A"

    # if the hash isn't in the known certificates list than it's authenticode
    # the only two types are authenticode and certificate however in practice only authenticode is used
    meta_data = {
        "flat_hash_sha256": row[0],
        "component": row[2] if row[2] != "" else None,
        "arch": convert_arch.get(row[3], row[3]),
        "partner": row[4],
        "type": "certificate"
        if authenticode_hash in KNOWN_CERTIFICATES
        else "authenticode",
        "cves": map_cve(row[5]),
        "date": row[6].replace("\n", ""),
        "authority": None,
        "links": [],
    }

    # Add links to the cve's if they exist
    for cve in meta_data["cves"].split("; "):
        if cve == "" or "XXXXX" in cve or "N/A" in cve:
            break

        meta_data["links"].append(f"https://nvd.nist.gov/vuln/detail/{cve}")

    # Add the authority if it exists
    # See https://oofhours.com/2021/01/19/uefi-secure-boot-who-controls-what-can-run/
    if meta_data["partner"] != "Unknown":
        if meta_data["partner"] == "Microsoft":
            meta_data["authority"] = "Microsoft Windows Production PCA 2011"
        else:
            meta_data["authority"] = "Microsoft Corporation UEFI CA 2011"

    return authenticode_hash, meta_data


def convert_uefi_org_revocation_file_to_dict(file: str) -> dict:
    """Converts the excel file to json.

    Args:
        file (str): Path to the excel or csv file

    Returns:
        dict: The data from the excel or csv file
    """
    data = {}

    if file.endswith(".xlsx"):
        wb = openpyxl.load_workbook(file)
        ws = wb.active
        for row in ws.iter_rows(min_row=2, values_only=True):
            authenticode_hash, meta_data = convert_row_to_metadata(row)
            data[authenticode_hash] = meta_data
    elif file.endswith(".csv"):
        with open(file, "r") as f:
            for i, row in enumerate(csv.reader(f)):
                if i == 0:
                    # Skip the header
                    continue
                authenticode_hash, meta_data = convert_row_to_metadata(row)
                data[authenticode_hash] = meta_data
    else:
        # This should never happen
        raise Exception("Unknown file type")

    return data


def convert_uefi_org_file(args: argparse.Namespace) -> None:
    """Converts the excel file to json.

    Args:
        args (argparse.Namespace): The arguments

    Returns:
        None
    """
    metadata = convert_uefi_org_revocation_file_to_dict(args.uefi_org_file)

    os.makedirs(os.path.dirname(args.output), exist_ok=True)

    with open(args.output, "w") as f:
        f.write(json.dumps(metadata, indent=4))

    logger.info("Wrote report to %s", args.output)


def generate_dbx_report(dbx_fs: BinaryIO, revocations: dict) -> dict:
    """Generates a report of the dbx.

    Args:
        dbx_fs (filestream): The dbx file system
        revocations (dict): The revocation list

    Returns:
        dict: The report
    """
    report = {
        "identified": {
            "dict": {},
            "total": 0,
            "note": "Represents all the hashes found in a systems dbx that match a provided revocation",
        },
        "missing_protections": {
            "dict": {},
            "total": 0,
            "note": "The remaining hashes in the provided revocation list that were not found in the system dbx",
        },
        "not_found": {
            "list": [],
            "total": 0,
            "note": "The hashes that were found in the dbx that are not in the provided revocation list",
        },
    }

    dbx = EfiSignatureDatabase(dbx_fs)
    for signature_list in dbx.EslList:
        for signature in signature_list.SignatureData_List:
            formatted_signature = ""
            for byte in signature.SignatureData:
                formatted_signature += f"{byte:02X}"

            # Is this in our revocation list?
            if formatted_signature in revocations:
                report["identified"]["dict"][formatted_signature] = revocations[
                    formatted_signature
                ]
                report["identified"]["dict"][formatted_signature]["count"] = 1

                # if we have identified the revocation, remove it from the list
                del revocations[formatted_signature]
            # have we already seen this signature?
            elif formatted_signature in report["identified"]["dict"]:
                report["identified"]["dict"][formatted_signature]["count"] += 1
            # otherwise, we have a signature that is not in our revocation list
            else:
                report["not_found"]["list"].append(formatted_signature)

    # Anything left in the revocation list is missing from the dbx
    for revocation in revocations:
        report["missing_protections"]["dict"][revocation] = revocations[revocation]

    # Add totals
    report["identified"]["total"] = len(report["identified"]["dict"])
    report["not_found"]["total"] = len(report["not_found"]["list"])
    report["missing_protections"]["total"] = len(
        report["missing_protections"]["dict"]
    )

    # Check for flat hashes
    for supposed_authenticode_hash in report["not_found"]["list"]:
        for supposed_flat_hash in revocations:
            if "flat_hash_sha256" not in revocations[supposed_flat_hash]:
                break

            if supposed_authenticode_hash == revocations[supposed_flat_hash]["flat_hash_sha256"]:
                logger.warning(
                    "This hash appears to be a flat sha256 hash: %s", supposed_authenticode_hash
                )

    return report


def filter_revocation_list_by_arch(revocations: dict, filter_by_arch: str=None) -> dict:
    """Filters the revocation list by architecture.

    Args:
        revocations (dict): The revocation list
        filter_by_arch (str): The architecture to filter by

    Returns:
        dict: The filtered revocation list
    """
    # Filter the revocations by arch if requested
    if filter_by_arch:
        filter_revocations = {}
        for revocation in revocations:
            if "arch" in revocations[revocation]:
                if revocations[revocation]["arch"] == filter_by_arch:
                    filter_revocations[revocation] = revocations[revocation]
        revocations = filter_revocations

    return revocations

###################################################################################################
# Classes
###################################################################################################


class FirmwareVariables(object):
    """Class to interact with firmware variables."""

    def __init__(self) -> None:
        """Constructor."""
        # enable required SeSystemEnvironmentPrivilege privilege
        privilege = win32security.LookupPrivilegeValue(
            None, "SeSystemEnvironmentPrivilege"
        )

        token = win32security.OpenProcessToken(
            win32process.GetCurrentProcess(),
            win32security.TOKEN_READ | win32security.TOKEN_ADJUST_PRIVILEGES,
        )

        win32security.AdjustTokenPrivileges(
            token, False, [(privilege, win32security.SE_PRIVILEGE_ENABLED)]
        )

        win32api.CloseHandle(token)

        try:
            self._GetFirmwareEnvironmentVariable = (
                KERNEL32.GetFirmwareEnvironmentVariableW
            )
            self._GetFirmwareEnvironmentVariable.restype = ctypes.c_int
            self._GetFirmwareEnvironmentVariable.argtypes = [
                ctypes.c_wchar_p,
                ctypes.c_wchar_p,
                ctypes.c_void_p,
                ctypes.c_int,
            ]
        except AttributeError:
            self._GetFirmwareEnvironmentVariable = None
            logger.warning("Get function doesn't exist")

    def get_variable(self, name: str, guid: str) -> bytes:
        """Gets a firmware variable.

        Args:
            name (str): Name of the variable to get
            guid (str): GUID of the variable to get

        Returns:
          The value of the variable
        """
        if self._GetFirmwareEnvironmentVariable is None:
            raise NotImplementedError(
                "GetFirmwareEnvironmentVariable is not implemented"
            )

        buffer = ctypes.create_string_buffer(EFI_VAR_MAX_BUFFER_SIZE)
        buffer_size = ctypes.c_int(EFI_VAR_MAX_BUFFER_SIZE)
        result = self._GetFirmwareEnvironmentVariable(name, guid, buffer, buffer_size)
        if result == 0:
            last_error = win32api.GetLastError()

            raise pywintypes.error(
                last_error,
                "GetFirmwareEnvironmentVariable",
                win32api.FormatMessage(last_error),
            )

        return buffer.raw[:result]


###################################################################################################
# main Functions
###################################################################################################


def get_secureboot_files(args: argparse.Namespace) -> int:
    """Gets the DBX from the system.

    Args:
        args (argparse.Namespace): the namespace from the ArgumentParser

    Returns:
        0 on success
    """
    if sys.platform != "win32":
        raise NotImplementedError("This function is only implemented for Windows")

    os.makedirs(args.output, exist_ok=True)

    # To make this cross platform, we need to make FirmwareVariables support other platforms
    fw_vars = FirmwareVariables()

    var = fw_vars.get_variable(
        args.secureboot_file, SECUREBOOT_FILES[args.secureboot_file]
    )

    output_file = os.path.join(args.output, f"{args.secureboot_file}.bin")

    with open(output_file, "wb") as f:
        f.write(var)

    with open(output_file, "rb") as f:
        EfiSignatureDatabase(f).Print()

    logger.info("Wrote %s", output_file)

    return 0


def parse_dbx(args: argparse.Namespace) -> int:
    """Parses the DBX.

    Args:
        args (argparse.Namespace): the namespace from the ArgumentParser

    Returns:
        0 on success
    """
    report = {}
    with open(args.revocations_file, "r", encoding="utf-8") as rev_fs:
        revocations = json.loads(rev_fs.read())

        with open(args.dbx_file, "rb") as dbx_fs:

            revocations = filter_revocation_list_by_arch(revocations, args.filter_by_arch)
            report = generate_dbx_report(dbx_fs, revocations)

    if args.format == "json":
        write_json_file(report, args.output + ".json")
    elif args.format == "xlsx":
        write_xlsx_file(report, args.output + ".xlsx")

    return 0


###################################################################################################
# Command Line Parsing Functions
###################################################################################################

def valid_file(param: str, valid_extensions: tuple=(".csv", ".xlsx")) -> str:
    """Checks if a file is valid.

    Args:
        param (str): the file to check
        valid_extensions (tuple): the valid extensions

    Returns:
        str: the file if it is valid
    """
    base, ext = os.path.splitext(param)
    if ext.lower() not in valid_extensions:
        raise argparse.ArgumentTypeError('File must be one of the following types: {}'.format(valid_extensions))
    return param


def setup_parse_dbx(subparsers: argparse._SubParsersAction) -> argparse._SubParsersAction:
    """Setup the parse_dbx subparser.

    Args:
        subparsers (argparse._SubParsersAction): the subparsers object from the ArgumentParser

    Returns:
        subparsers (argparse._SubParsersAction): the subparsers object from the ArgumentParser
    """
    parser = subparsers.add_parser("parse_dbx")
    parser.set_defaults(function=parse_dbx)

    parser.add_argument("dbx_file", help="Input file to read the DBX from")

    parser.add_argument(
        "revocations_file",
        help="JSON file containing the revocations to check against the DBX",
    )

    parser.add_argument(
        "--filter-by-arch",
        help="Filter the output to only include the specified arch type",
        choices=["x86", "x86_64", "arm", "arm64"],
        default=None,
    )

    parser.add_argument(
        "--output",
        help="Output file to write the dbx contents to"
        + " (note: extension will be based on the format)",
        default=os.path.join(DEFAULT_OUTPUT_FOLDER, "dbx_report"),
    )

    parser.add_argument(
        "--format",
        help="Format of the revocations file",
        choices=["json", "xlsx"],
        default="json",
    )

    return subparsers

def setup_parse_uefi_org_files(subparsers: argparse._SubParsersAction) -> argparse._SubParsersAction:
    """Setup the parse_uefi_org_files subparser.

    Args:
        subparsers (argparse._SubParsersAction): the subparsers object from the ArgumentParser

    Returns:
        subparsers (argparse._SubParsersAction): the subparsers object from the ArgumentParser
    """
    parser = subparsers.add_parser("convert_file")
    parser.set_defaults(function=convert_uefi_org_file)

    parser.add_argument(
        "uefi_org_file",
        type=valid_file,
        help="The csv or excel file from uefi.org to parse, this will provide metadata about the signatures",
    )

    parser.add_argument(
        "--output",
        help="Output file to write the uefi org revocations to",
        default=os.path.join(DEFAULT_OUTPUT_FOLDER, "uefi_org_revocations.json"),
    )

    return subparsers


def setup_get_secureboot_files(subparsers: argparse._SubParsersAction) -> argparse._SubParsersAction:
    """Setup the get_dbx subparser.

    Args:
        subparsers (argparse._SubParsersAction): the subparsers object from the ArgumentParser

    Returns:
        subparsers (argparse._SubParsersAction): the subparsers object from the ArgumentParser
    """
    parser = subparsers.add_parser("get")
    parser.set_defaults(function=get_secureboot_files)

    parser.add_argument(
        "secureboot_file",
        help="The secureboot file to get from UEFI",
        choices=SECUREBOOT_FILES.keys(),
    )

    parser.add_argument(
        "--output",
        help="Output folder to write the secureboot variables to",
        default=DEFAULT_OUTPUT_FOLDER,
    )

    return subparsers


def parse_args() -> argparse.Namespace:
    """Parses arguments from the command line."""
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()
    subparsers = setup_get_secureboot_files(subparsers)
    subparsers = setup_parse_uefi_org_files(subparsers)
    subparsers = setup_parse_dbx(subparsers)

    args = parser.parse_args()

    if not hasattr(args, "function"):
        parser.print_help(sys.stderr)
        sys.exit(1)

    return args


def main() -> None:
    """Main function."""
    args = parse_args()

    # Raise a SystemExit exception with the exit status
    sys.exit(args.function(args))


if __name__ == "__main__":
    main()



# @file versioninfo_tool.py
# This module contains the CLI interface for easily creating VERSIONINFO resources
# from json files.
#
##
# Copyright (C) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A Command-line tool to generate VERSIONINFO resource files.

Versioninfo Tool is a command-line tool to assist in generating VERSIONINFO
resource files for use with a Resource Compiler. It takes a JSON representing
versioning info and produces a resource file that once compiled will create a
standard resource section.
"""

import argparse
import json
import logging
import os
import sys
import typing

from edk2toolext.versioninfo.versioninfo_helper import PEObject, VERSIONINFOGenerator

TOOL_VERSION = "0.7.0"  # please change this as the tool is updated.

TOOL_DESCRIPTION = """
Versioninfo Tool is a command-line tool to assist in generating VERSIONINFO
resource files for use with a Resource Compiler. It takes a JSON representing
versioning info and produces a resource file that once compiled will create a
standard resource section.

!!! Warning - BETA Feature
This tool is still in early development and may change with
little regard for backward compatibility.

Version: %s

An example to encode json to rc file might look like:
%s -e /path/to/version.JSON /path/to/output

An example to decode a binary efi file and output the rsrc in json might look like:
%s -d /path/to/app.efi /path/to/output.JSON
""" % (TOOL_VERSION, os.path.basename(sys.argv[0]), os.path.basename(sys.argv[0]))


def get_cli_options(args:typing.Sequence[str]=None) -> argparse.Namespace:
    """Parse options from the command line.

    Will parse the primary options from the command line. If provided, will take the options as
    an array in the first parameter
    """
    parser = argparse.ArgumentParser(description=TOOL_DESCRIPTION, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('input_file', type=str,
                        help='a filesystem path to a json/PE file to load')
    parser.add_argument('output_file', type=str,
                        help='a filesystem path to the output file. if file does not exist, entire directory path will be created. if file does exist, contents will be overwritten')  # noqa

    command_group = parser.add_mutually_exclusive_group()
    command_group.add_argument('-e', '--encode', action='store_const', const='e', dest='mode',
                               help='(default) outputs VERSIONINFO.rc of given json file')
    command_group.add_argument('-d', '--dump', action='store_const', dest='mode', const='d',
                               help='outputs json file of VERSIONINFO given PE file')
    parser.set_defaults(mode='e')
    return parser.parse_args(args=args)


def decode_version_info(input_file: str) -> dict:
    """Takes in a PE file (input_file) and returns the version dictionary.

    Args:
        input_file (str): path to a PE file

    Returns:
        (dict): version dictionary
    """
    if not os.path.exists(input_file):
        raise FileNotFoundError(input_file)
    pe = PEObject(input_file)
    return pe.get_version_dict()


def decode_version_info_dump_json(input_file: str, output_file: str) -> bool:
    """Takes in a PE file (input_file) and dumps the output as json to (output_file).

    Args:
        input_file (str): Path to a PE file
        output_file (str): Path to json dump

    Returns:
        (bool): result of parsing and dumping
    """
    version_info = decode_version_info(input_file)
    if not version_info:
        logging.error(f"{input_file} doesn't have version info.")
        return False
    with open(output_file, "w") as out:
        json.dump(version_info, out)
    return True


def encode_version_info(input_file: str) -> VERSIONINFOGenerator:
    """Takes in a JSON file (inputfile) and returns an object.

    Args:
        input_file (str): path to JSON File

    Returns:
        (VERSIONINFOGenerator): VERSIONINFO rc file Generator object
    """
    if not os.path.exists(input_file):
        raise FileNotFoundError(input_file)
    return VERSIONINFOGenerator(input_file)


def encode_version_info_dump_rc(input_file: str, output_file: str) -> bool:
    """Takes in a JSON file (input_file) and outputs an RC file(output_file).

    Args:
        input_file (str): path to JSON file
        output_file (str): path to RC file

    Returns:
        (bool): Result of writing to file
    """
    return encode_version_info(input_file).write(output_file, TOOL_VERSION)


def main() -> None:
    """Parse args, executes versioninfo_tool."""
    logging.getLogger().addHandler(logging.StreamHandler())
    args = get_cli_options()
    if not os.path.isfile(args.input_file):
        logging.error("Could not find " + args.input_file)
        sys.exit(1)

    if args.mode == 'd':
        # we need to dump
        if not decode_version_info_dump_json(args.input_file, args.output_file):
            sys.exit(1)
    elif args.mode == 'e':
        # we need to encode
        if not encode_version_info_dump_rc(args.input_file, args.output_file):
            sys.exit(1)
    else:
        # unknown mode
        logging.error(f"Unknown mode: {args.mode}")
        sys.exit(1)


if __name__ == '__main__':
    main()



# @file base_abstract_invocable
# Base class for an Invocable. Loads environment before calling subclass.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

"""The Base abstract Invocable that all other invocables should inherit from."""
import logging
import os
import sys
from datetime import datetime
from typing import Optional

from edk2toolext import edk2_logging
from edk2toolext.environment import plugin_manager, self_describing_environment
from edk2toolext.environment.plugintypes.uefi_helper_plugin import HelperFunctions


class BaseAbstractInvocable(object):
    """The Abstract Invocable.

    The base abstract invocable that other invocables should inherit from.
    Provides base functionality to configure logging and the environment.

    Attributes:
        log_filename (str): logfile path
        plugin_manager (plugin_manager.PluginManager): the plugin manager
        helper (HelperFunctions): container for all helper functions
    """
    def __init__(self) -> None:
        """Init the Invocable."""
        self.log_filename = None
        return

    def ParseCommandLineOptions(self) -> None:
        """Parse command line arguments.

        !!! tip
            Required Override in a subclass
        """
        raise NotImplementedError()

    def GetWorkspaceRoot(self) -> str:
        """Return the workspace root for initializing the Self Describing Environment.

        !!! tip
            Required Override in a subclass

        Returns:
            (str): absolute path to workspace root

        """
        raise NotImplementedError()

    def GetActiveScopes(self) -> tuple:
        """Return tuple containing scopes that should be active for this process.

        !!! tip
            Required Override in a subclass

        !!! warning
            A single scope should end in a comma i.e. (scope,)

        Returns:
            (Tuple): scopes
        """
        raise NotImplementedError()

    def GetSkippedDirectories(self) -> tuple:
        """Return tuple containing workspace-relative directory paths that should be skipped for processing.

        !!! tip
            Optional Override in a subclass

        !!! warning
            A single directory should end with a comma i.e. (dir,)

        Returns:
            (Tuple): directories
        """
        return ()

    def GetLoggingLevel(self, loggerType: str) -> Optional[int]:
        """Get the logging level depending on logger type.

        !!! tip
            Required Override in a subclass

        Args:
            loggerType (str): type of logger being logged to

        Returns:
            (int): The logging level
            (None): No logging of this type

        !!! note "loggerType possible values"
            "base": lowest logging level supported

            "con": logs to screen

            "txt": logs to plain text file
        """
        raise NotImplementedError()

    def GetLoggingFolderRelativeToRoot(self) -> str:
        """Return the path to a directory to hold all log files.

        !!! hint
            Required Override in a subclass

        Returns:
            (str): path to the directory
        """
        raise NotImplementedError()

    def InputParametersConfiguredCallback(self) -> int:
        """A Callback once all input parameters are collected.

        !!! hint
            Optional override in subclass
        """

    def GetVerifyCheckRequired(self) -> bool:
        """Will call self_describing_environment.VerifyEnvironment if this returns True.

        !!! hint
            Optional override in a subclass

        Returns:
            (bool): whether verify check is required or not
        """
        return True

    def GetLoggingFileName(self, loggerType: str) -> Optional[str]:
        """Get the logging File name to provide file name customization.

        !!! hint
            Required Override this in a subclass

        Args:
            loggerType (str): values can be base, con, txt, md. See hint below

        Returns:
            (str): filename
            (None): No logging file should be created

        !!! note "loggerType possible values"
            "base": lowest logging level supported

            "con": logs to screen

            "txt": logs to plain text file
        """
        raise NotImplementedError()

    def Go(self) -> None:
        """Main function to run.

        Main function to run after the environment and logging has been configured.

        !!! tip
            Required Override in a subclass
        """
        raise NotImplementedError()

    def ConfigureLogging(self) -> None:
        """Sets up the logging.

        !!! tip
            Optional override in a subclass if new behavior is needed
        """
        logger = logging.getLogger('')
        logger.setLevel(self.GetLoggingLevel("base"))

        # Adjust console mode depending on mode.
        edk2_logging.setup_section_level()

        edk2_logging.setup_console_logging(self.GetLoggingLevel("con"))

        log_directory = os.path.join(self.GetWorkspaceRoot(), self.GetLoggingFolderRelativeToRoot())

        txtlogfile = self.GetLoggingLevel("txt")
        if (txtlogfile is not None):
            logfile, filelogger = edk2_logging.setup_txt_logger(log_directory,
                                                                self.GetLoggingFileName("txt"),
                                                                txtlogfile)
            self.log_filename = logfile

        logging.info("Log Started: " + datetime.strftime(datetime.now(), "%A, %B %d, %Y %I:%M%p"))

    def Invoke(self) -> None:
        """Main process function to configure logging and the environment.

        !!! danger
            Do not override this method
        """
        self.ParseCommandLineOptions()
        self.ConfigureLogging()
        self.InputParametersConfiguredCallback()

        logging.log(edk2_logging.SECTION, "Init Self Describing Environment")

        #
        # Next, get the environment set up.
        #
        (build_env, shell_env) = self_describing_environment.BootstrapEnvironment(
            self.GetWorkspaceRoot(), self.GetActiveScopes(), self.GetSkippedDirectories())

        # Make sure the environment verifies IF it is required for this invocation
        if self.GetVerifyCheckRequired() and not self_describing_environment.VerifyEnvironment(
                self.GetWorkspaceRoot(), self.GetActiveScopes(), self.GetSkippedDirectories()):
            raise RuntimeError("External Dependencies in the environment are out of date. "
                               "Consider running stuart_update to possibly resolve this issue.")

        # Load plugins
        logging.log(edk2_logging.SECTION, "Loading Plugins")

        self.plugin_manager = plugin_manager.PluginManager()
        failedPlugins = self.plugin_manager.SetListOfEnvironmentDescriptors(
            build_env.plugins)
        if failedPlugins:
            logging.critical("One or more plugins failed to load. Halting build.")
            for a in failedPlugins:
                logging.error("Failed Plugin: {0}".format(a["name"]))
            raise Exception("One or more plugins failed to load.")

        self.helper = HelperFunctions()
        if (self.helper.LoadFromPluginManager(self.plugin_manager) > 0):
            raise Exception("One or more helper plugins failed to load.")

        logging.log(edk2_logging.SECTION, "Start Invocable Tool")
        retcode = self.Go()
        logging.log(edk2_logging.SECTION, "Summary")
        if (retcode != 0):
            logging.error("Error")
        else:
            edk2_logging.log_progress("Success")

        logging.shutdown()
        sys.exit(retcode)



# @file coverage_report.py
# A report that re-organizes a cobertura.xml by INF.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A report that re-organizes a cobertura.xml by INF."""
import fnmatch
import logging
import os
import re
import xml.dom.minidom as minidom
import xml.etree.ElementTree as ET
from argparse import Action, ArgumentParser, Namespace
from pathlib import Path
from typing import Optional, Sequence, Tuple

from edk2toollib.database import Edk2DB, Environment, Fv, Inf, InstancedInf, Package, Session, Source, Value
from edk2toollib.uefi.edk2.path_utilities import Edk2Path
from sqlalchemy import func
from sqlalchemy.orm import aliased

from edk2toolext.environment.reporttypes.base_report import Report


class SplitCommaAction(Action):
    """A Custom action similar to append, but will split the input string on commas first."""
    def __call__(
            self, parser: ArgumentParser,
            namespace: Namespace,
            values: str | Sequence[str],
            option_string: Optional[str] = None,
            ) -> None:
       """Command entry."""
       setattr(namespace, self.dest, getattr(namespace, self.dest, []) + values.split(','))

class CoverageReport(Report):
    """A report that re-organizes a cobertura.xml by INF.

    This report will supports two modes, by-package and by-platform. By-package will only include coverage data for
    files in the specified edk2 packages. By-platform will only include coverage data for files used to build the
    specified platform dsc.
    """
    def report_info(self) -> Tuple[str, str]:
        """Returns the report standard information.

        Returns:
            (str, str): A tuple of (name, description)
        """
        return ("coverage", "Reorganizes an xml coverage report by INF rather than executable. Filters results based "
                "on --by-package or --by-platform flags.")

    def add_cli_options(self, parserobj: ArgumentParser) -> None:
        """Configure command line arguments for this report."""
        # Group 2 - Calculate coverage only on files used by a specific platform
        group = parserobj.add_argument_group("Coverage by platform options")
        group.add_argument("--by-platform", action="store_true", dest="by_platform", default=False,
                           help="Filters test coverage to all files used to build the specified platform package.")
        group.add_argument("-d", "--dsc", "--DSC", dest="dsc",
                           help="Edk2 relative path the ACTIVE_PLATFORM DSC file.")

        # Group 3 - Run either by-platform or by-package with a FULL report
        group = parserobj.add_argument_group("Full Report")
        group.add_argument("--full", action="store_true", dest="full", default=False,
                           help="Include all files in the report, not just those with coverage data. Requires pygount.")
        group.add_argument("-ws", "--workspace", "--Workspace", "--WORKSPACE", dest="workspace",
                               help="The Workspace root associated with the xml argument.", default=".")

        # Other args
        parserobj.add_argument(dest="xml", action="store", help="The path to the XML file parse.")
        parserobj.add_argument("-o", "--output", "--Output", "--OUTPUT", dest="output", default="Coverage.xml",
                               help="The path to the output XML file.", action="store")
        parserobj.add_argument("-e", "--exclude", "--Exclude", "--EXCLUDE", dest="exclude",
                               action=SplitCommaAction, default=[],
                               help="Package path relative paths or file (.txt). Globbing is supported. Can be "
                               "specified multiple times")
        parserobj.add_argument("-p", "--package", "--Package", "--PACKAGE", dest="package_list",
                           action=SplitCommaAction, default=[],
                           help="The package to include in the report. Can be specified multiple times.")
        parserobj.add_argument("--flatten", action="store_true", dest="flatten", default=False,
                              help="Flatten the report to only source files. This removes duplicate files that are in "
                              "multiple INFs.")
        group = parserobj.add_argument_group("Deprecated Options")
        group.add_argument("--by-package", action="store_true", dest="by_package", default=False,
                           help="Filters test coverage to only files in the specified packages(s)")

    def run_report(self, db: Edk2DB, args: Namespace) -> None:
        """Generate the Coverage report."""
        self.args = args

        if self.args.full:
            self.verify_pygount()

        self.update_excluded_files()

        with db.session() as session:
            package_list = self.args.package_list or [pkg.name for pkg in session.query(Package).all()]
            logging.info(f"Packages requested: {', '.join(package_list)}")
            if args.by_platform:
                logging.info("Organizing coverage report by Platform.")
                return self.run_by_platform(session, package_list)
            if args.by_package:
                logging.warning(
                    "The --by-package flag is deprecated and will be removed in a future release."
                    " by-package is now the default behavior and overriden with by-platform."
                )
            return self.run_by_package(session, package_list)

    def run_by_platform(self, session: Session, package_list: list) -> None:
        """Runs the report, only adding coverage data for source files used to build the platform.

        Args:
            session (Session): The session associated with the database
            package_list (list): The list of packages to filter the results by

        Returns:
            (bool): True if the report was successful, False otherwise.
        """
        # Verify valid ACTIVE_PLATFORM
        dsc = self.args.dsc
        if not dsc:
            logging.error("No ACTIVE_PLATFORM dsc file specified. It should be edk2 package path relative.")
            return -1
        logging.info(f"ACTIVE_PLATFORM requested: {dsc}")

        result = (
            session
                .query(Environment)
                .filter(Environment.values.any(key="ACTIVE_PLATFORM", value=dsc))
                .order_by(Environment.date.desc())
                .first()
        )

        if result is None:
            logging.error(f"Could not locate an ACTIVE_PLATFORM containing the provided package: {dsc}")
            return -1
        env_id = result.id

        # Build source / coverage association dictionary
        coverage_files = self.build_source_coverage_dictionary(self.args.xml, package_list)

        # Build inf / source association dictionary
        inf_alias = aliased(InstancedInf)
        inf_list = (
            session
                .query(inf_alias)
                .join(Fv.infs)
                .join(inf_alias, InstancedInf.path == inf_alias.component)
                .filter(Fv.env == env_id)
                .filter(inf_alias.env == env_id)
                .filter(InstancedInf.env == env_id)
                .group_by(inf_alias.path)
                .distinct(inf_alias.path)
                .all()
        )
        data = [
            (inf.path, source.path) for inf in inf_list for source in inf.sources if source.path.lower().endswith(".c")
        ]

        package_files = self.build_inf_source_dictionary(data, package_list)

        # Build the report
        return self.build_report(session, env_id, coverage_files, package_files)

    def run_by_package(self, session: Session, package_list: list) -> bool:
        """Runs the report, only adding coverage data for source files in the specified packages.

        Args:
            session (Session): The session associated with the database
            package_list (list): The list of packages to filter the results by

        Returns:
            (bool): True if the report was successful, False otherwise.
        """
        # Get env_id
        env_id, = session.query(Environment.id).order_by(Environment.date.desc()).first()

        # Build source / coverage association dictionary
        coverage_files = self.build_source_coverage_dictionary(self.args.xml, package_list)

        # Build inf / source association dictionary
        data = (
            session
                .query(Inf.path, Source.path)
                .join(Inf.sources)
                .filter(func.lower(Source.path).endswith('.c'))
                .group_by(Inf.path, Source.path)
                .distinct(Inf.path, Source.path)
                .all()
        )
        package_files = self.build_inf_source_dictionary(data, package_list)

        # Build the report
        return self.build_report(session, env_id, coverage_files, package_files)

    def build_source_coverage_dictionary(self, xml_path: str, package_list: list) -> dict:
        """Builds a dictionary of source files and their coverage data.

        Args:
            xml_path (str): path to xml file storing coverage data
            package_list (list): list of packages that a file must be in to be included in the report

        Returns:
            dict[str, ET.Element]: A dictionary of source files and their coverage data.
        """
        tree = ET.parse(xml_path)
        regex = re.compile('|'.join(map(re.escape, package_list)))
        file_dict = {}
        for file in tree.iter("class"):
            # Add the file results if they do not exist
            filename = file.get("filename")
            match = regex.search(filename)

            # Skip the file if it is not in a package the user care about (-p)
            if not match:
                continue

            path = Path(filename[match.start():]).as_posix()
            if path not in file_dict:
                file.attrib["filename"] = path
                file.attrib["name"] = "\\".join(Path(path).parts)
                file_dict[path] = file

            # Merge the file results
            else:
                to_update = file_dict[path]

                # Merge lines first
                for line in file.find("lines").iter():
                    line_number = line.attrib.get("number")
                    if not line_number:
                        continue
                    match = to_update.find("lines").find(f".//line[@number='{line_number}']")
                    match.set("hits", str(int(match.attrib.get("hits")) + int(line.attrib.get("hits"))))
        return file_dict

    def build_inf_source_dictionary(self, data: dict, package_list: list) -> dict:
        """Builds a dictionary of INFs and the source files they use.

        Args:
            data (dict): The data to build the dictionary from
            package_list (list): The packages to filter the results by

        Returns:
            dict[str, list[str]]: A dictionary of INFs and the source files they use.
        """
        entry_dict = {}

        for inf, source in data:
            if not any(inf.startswith(pkg) for pkg in package_list):
                continue
            if inf not in entry_dict:
                entry_dict[inf] = [source]
            else:
                entry_dict[inf].append(source)
        return entry_dict

    def build_report(self, session: Session, env_id: int, source_coverage_dict: dict, inf_source_dict: dict) -> None:
        """Builds the report.

        For each source file in each INF in the inf_source dictionary, look to see if there is coverage data for it in
        the source_coverage dictionary. If it exists, insert it into the new report. Writes the final report to the
        specified file.
        """
        pp_list = session.query(Value).filter_by(env_id=env_id, key="PACKAGES_PATH").one().value
        pp_list = pp_list.split(os.pathsep)
        edk2path = Edk2Path(self.args.workspace, pp_list)

        root = ET.Element("coverage")
        sources = ET.SubElement(root, "sources")
        # Set the sources so that reports can find the right paths.
        source = ET.SubElement(sources, "source")
        source.text = self.args.workspace
        for pp in pp_list:
            source = ET.SubElement(sources, "source")
            source.text = str(Path(self.args.workspace, pp))

        packages = ET.SubElement(root, "packages")
        for path, source_list in inf_source_dict.items():
            if fnmatch.fnmatch(path, "*Test*"):
                continue
            if not source_list:
                continue
            inf = ET.SubElement(packages, "package", path=path, name=Path(path).name)
            classes = ET.SubElement(inf, "classes")

            for source in source_list:
                # Check if the file should be excluded
                exclude_file = False
                for pattern in self.args.exclude:
                    if fnmatch.fnmatch(source, pattern):
                        logging.debug(f'{source} excluded due to {pattern}')
                        exclude_file = True
                        break
                if exclude_file:
                    continue

                match = next((key for key in source_coverage_dict.keys() if Path(source).is_relative_to(key)), None)
                if match is not None:
                    classes.append(source_coverage_dict[match])
                elif self.args.full:
                    classes.append(self.create_source_xml(source, edk2path))

        # Flaten the report to only source files, removing duplicates from INFs.
        if self.args.flatten:
            root = self.flatten_report(root, edk2path)

        xml_string = ET.tostring(root, "utf-8")
        dom = minidom.parseString(xml_string)
        dt = minidom.getDOMImplementation('').createDocumentType(
            'coverage', None, "http://cobertura.sourceforge.net/xml/coverage-04.dtd")
        dom.insertBefore(dt, dom.documentElement)
        p = Path(self.args.output)
        p.unlink(missing_ok=True)
        with open(p, 'wb') as f:
            f.write(dom.toprettyxml(encoding="utf-8", indent="  "))
        logging.info(f"Coverage xml data written to {p}")

    def update_excluded_files(self) -> None:
        """Replaces any files in the exclude list with their contents."""
        temporary_list = []
        for pattern in self.args.exclude:
            if Path(pattern).exists() and Path(pattern).suffix == ".txt":
                with open(pattern, "r") as f:
                    temporary_list.extend(f.read().splitlines())
            else:
                temporary_list.append(pattern)
        self.args.exclude = temporary_list

    def create_source_xml(self, source_path: str, edk2path: Edk2Path) -> ET:
        """Parses the source file and creates a coverage 'lines' xml element for it."""
        from pygount import SourceAnalysis
        full_path = edk2path.GetAbsolutePathOnThisSystemFromEdk2RelativePath(source_path)
        code_count = SourceAnalysis.from_file(full_path, "_").code_count
        file_xml = ET.Element("class", name="\\".join(Path(source_path).parts), filename=source_path)
        lines_xml = ET.Element("lines")

        for i in range(1, code_count + 1):
            lines_xml.append(ET.Element("line", number=str(i), hits="0"))
        file_xml.append(lines_xml)
        return file_xml

    def flatten_report(self, root: ET.Element, edk2path: Edk2Path) -> ET.Element:
        """Flattens the report to only source files, removing the INF layer and duplicate source files."""
        class_list = ET.Element("classes")

        class_dict = {}
        for class_element in root.iter("class"):
            filename = class_element.get('filename')
            filename = "\\".join(Path(filename).parts)
            class_element.set("name", filename)
            class_dict[filename] = class_element

        for class_element in class_dict.values():
            class_list.append(class_element)

        package_element = ET.Element("package", name = "All Source")
        package_element.append(class_list)

        packages = root.find('.//packages')
        packages.clear()
        packages.append(package_element)
        return root

    def verify_pygount(self) -> None:
        """Verify that pygount is installed."""
        try:
            from pygount import SourceAnalysis  # noqa: F401
        except ImportError as e:
            print(e)
            print("WARNING: This report requires pip modules not installed with edk2-pytool-extensions:")
            print("  Run the following command: `pip install pygount`")
            exit(-1)



# @file edk2_setup
# updates submodules listed as Required Submodules in Config file.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code that updates required submodules.

Contains a SetupSettingsManager that must be subclassed in a build settings
file. This provides platform specific information to Edk2PlatformSetup invocable
while allowing the invocable itself to remain platform agnostic.
"""
import argparse
import logging
import os
from typing import Iterable

from edk2toollib.utility_functions import GetHostInfo

from edk2toolext import edk2_logging
from edk2toolext.environment import version_aggregator
from edk2toolext.environment.repo_resolver import (
    GitCommandError,
    InvalidGitRepositoryError,
    clean,
    repo_details,
    submodule_clean,
    submodule_resolve,
)
from edk2toolext.invocables.edk2_multipkg_aware_invocable import (
    Edk2MultiPkgAwareInvocable,
    MultiPkgAwareSettingsInterface,
)


class RequiredSubmodule():
    """A class containing information about a git submodule."""
    def __init__(self, path: str, recursive: bool = True) -> None:
        """Object to hold necessary information for resolving submodules.

        Args:
            path (str): workspace relative path to submodule that must be
                synchronized and updated
            recursive (bool): if recursion should be used in this submodule
        """
        self.path = path
        self.recursive = recursive


class SetupSettingsManager(MultiPkgAwareSettingsInterface):
    """Platform specific settings for Edk2PlatformSetup.

    Provides information necessary for `stuart_setup.exe` or `edk2_setup.py`
    to successfully execute for a given platform.

    !!! example "Example of Overriding SetupSettingsManager"
        ```python
        from edk2toolext.invocables.edk2_setup import SetupSettingsManager, RequiredSubmodule
        class PlatformManager(SetupSettingsManager):
            def GetRequiredSubmodules(self) -> Iterable[RequiredSubmodule]:
                return [RequiredSubmodule('Common/MU', True)]
        ```
    """

    def GetRequiredSubmodules(self) -> Iterable[RequiredSubmodule]:
        """Provides a list of required git submodules.

        These submodules are those that must be setup for the platform
        to successfully build.

        !!! tip
            Optional Override in a subclass

        Returns:
            A Iterable of required submodules, or an empty Iterable
        """
        return []


class Edk2PlatformSetup(Edk2MultiPkgAwareInvocable):
    """Invocable that updates git submodules listed in RequiredSubmodules."""

    def AddCommandLineOptions(self, parserObj: argparse.ArgumentParser) -> None:
        """Adds command line options to the argparser."""
        parserObj.add_argument('--force', '--FORCE', '--Force', dest="force", action='store_true', default=False)
        parserObj.add_argument('--omnicache', '--OMNICACHE', '--Omnicache', dest='omnicache_path',
                               default=os.environ.get('OMNICACHE_PATH'))

        super().AddCommandLineOptions(parserObj)

    def RetrieveCommandLineOptions(self, args: argparse.Namespace) -> None:
        """Retrieve command line options from the argparser."""
        self.force_it = args.force
        self.omnicache_path = args.omnicache_path
        if (self.omnicache_path is not None) and (not os.path.exists(self.omnicache_path)):
            logging.warning(f"Omnicache path set to invalid path: {args.omnicache_path}")
            self.omnicache_path = None

        super().RetrieveCommandLineOptions(args)

    def GetVerifyCheckRequired(self) -> bool:
        """Will not call self_describing_environment.VerifyEnvironment because it hasn't been set up yet."""
        return False

    def GetSettingsClass(self) -> type:
        """Returns the SetupSettingsManager class.

        !!! warning
            SetupSettingsManager must be subclassed in your platform settings file.
        """
        return SetupSettingsManager

    def GetLoggingFileName(self, loggerType: str) -> str:
        """Returns the filename (SETUPLOG) of where the logs for the Edk2CiBuild invocable are stored in."""
        return "SETUPLOG"

    def Go(self) -> int:
        """Executes the core functionality of the Edk2PlatformSetup invocable."""
        required_submodules = self.PlatformSettings.GetRequiredSubmodules()

        # Return clear error if submodules are a windows format on a non-windows system
        # This will cause errors with git commands and produces ugly errors.
        if GetHostInfo().os != "Windows":
            for submodule in required_submodules:
                if submodule.path.find("\\") != -1:
                    logging.error("Windows Path format detected on a non-Windows system. This is not supported.")
                    logging.error(f"    Path: {submodule.path}")
                    logging.error("    Defined at: GetRequiredSubmodules()")
                    return -1

        workspace_path = self.GetWorkspaceRoot()

        details = repo_details(workspace_path)
        git_version = details['GitVersion']

        version_aggregator.GetVersionAggregator().ReportVersion("Git",
                                                                git_version,
                                                                version_aggregator.VersionTypes.TOOL)

        # Pre-setup cleaning if "--force" is specified.
        if self.force_it:
            try:
                # Clean the workspace
                edk2_logging.log_progress('## Cleaning the root repo')
                clean(workspace_path, ignore_files=[f'Build/{self.GetLoggingFileName("txt")}.txt'])
                edk2_logging.log_progress("Done.\n")
            except InvalidGitRepositoryError:
                logging.error(f"Error when trying to clean {workspace_path}")
                logging.error(f"Invalid Git Repository at {workspace_path}")
                return -1

            # Clean the submodules
            for required_submodule in required_submodules:
                try:
                    submodule_path = os.path.join(workspace_path, required_submodule.path)
                    edk2_logging.log_progress(f'## Cleaning Git Submodule: {required_submodule.path}')
                    submodule_clean(workspace_path, required_submodule)
                    edk2_logging.log_progress('## Done.\n')
                except InvalidGitRepositoryError:
                    logging.error(f"Error when trying to clean {submodule_path}")
                    logging.error(f"Invalid Git Repository at {submodule_path}")
                    return -1
                except ValueError as e:
                    logging.error(f"Error when trying to clean {submodule_path}")
                    logging.error(e)
                    return -1

        # Resolve all of the submodules to the specifed branch and commit. i.e. sync, then update
        for submodule in required_submodules:
            edk2_logging.log_progress(f'## Resolving Git Submodule: {submodule.path}')
            submodule_details = repo_details(os.path.join(workspace_path, submodule.path))

            # Don't update a dirty submodule unless we are forcing.
            if submodule_details['Dirty'] and not self.force_it:
                logging.info('-- NOTE: Submodule currently exists and appears to have local changes!')
                logging.info('-- Skipping fetch!')
                logging.log(edk2_logging.get_progress_level(), "Done.\n")
                continue
            try:
                # Sync, then Update & Init the submodule
                submodule_resolve(workspace_path, submodule, omnicache_path=self.omnicache_path)
                edk2_logging.log_progress('## Done.\n')
            except InvalidGitRepositoryError:
                logging.error(f"Error when trying to resolve {submodule.path}")
                logging.error(f"Invalid Git Repository at {submodule.path}")
                return -1
            except GitCommandError as e:
                logging.error(f"Error when trying to resolve {submodule.path}")
                logging.error(e)
                return -1

        return 0


def main() -> None:
    """Entry point to invoke Edk2PlatformSetup."""
    Edk2PlatformSetup().Invoke()



# @file nuget_dependency.py
# This module implements ExternalDependency for NuGet packages.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""An ExternalDependency subclass able to download from NuGet."""
import logging
import os
import shutil
from io import StringIO
from typing import Optional

import semantic_version
from edk2toollib.utility_functions import GetHostInfo, RemoveTree, RunCmd

from edk2toolext.bin.nuget import DownloadNuget
from edk2toolext.environment.external_dependency import ExternalDependency


class NugetDependency(ExternalDependency):
    """An ExternalDependency subclass able to download from NuGet.

    Attributes:
        source (str): Source of the nuget dependency.
        version (str): Version of the web dependency.

    !!! tip
        The attributes are what must be described in the ext_dep yaml file!
    """
    TypeString = "nuget"

    # Env variable name for path to folder containing NuGet.exe
    NUGET_ENV_VAR_NAME = "NUGET_PATH"

    def __init__(self, descriptor: dict) -> None:
        """Inits a nuget dependency based off the provided descriptor."""
        super().__init__(descriptor)
        self.nuget_cache_path = None

    @classmethod
    def GetNugetCmd(cls: 'NugetDependency') -> list[str]:
        """Appends mono to the command and resolves the full path of the exe for mono.

        Used to add nuget support on posix platforms.
        https://docs.microsoft.com/en-us/nuget/install-nuget-client-tools

        !!! note
            Strings returned might not be pathlike given they may be quoted
            for use on the command line.

        Returns:
            (list): ["nuget.exe"] or ["mono", "/PATH/TO/nuget.exe"]
            (None): none was found
        """
        cmd = []
        if GetHostInfo().os != "Windows":
            cmd += ["mono"]

        nuget_path = os.getenv(cls.NUGET_ENV_VAR_NAME)
        if nuget_path is not None:
            nuget_path = os.path.join(nuget_path, "NuGet.exe")
            if not os.path.isfile(nuget_path):
                logging.info(f'{cls.NUGET_ENV_VAR_NAME} set, but did not exist. Attempting to download.')
                DownloadNuget(nuget_path)
        else:
            nuget_path = DownloadNuget()

        if not os.path.isfile(nuget_path):
            logging.error("We weren't able to find or download NuGet!")
            return None

        # Make sure quoted string if it has spaces
        if " " in nuget_path.strip():
            nuget_path = '"' + nuget_path + '"'

        cmd += [nuget_path]

        return cmd

    @staticmethod
    def normalize_version(version: str, nuget_name: Optional[str]="") -> str:
        """Normalizes the version as NuGet versioning diverges from Semantic Versioning.

        https://learn.microsoft.com/en-us/nuget/concepts/package-versioning#where-nugetversion-diverges-from-semantic-versioning

        These cases will be handled befpre a Semantic Version Compatible" set
        of data is passed to the Semantic Version checker.
        """
        # 1. NuGetVersion requires the major segment to be defined
        if not version:
            raise ValueError("String is empty. At least major version is "
                             "required.")

        # 2. NuGetVersion uses case insensitive string comparisons for
        #    pre-release components
        reformed_ver = version.strip().lower()

        tag = None
        parts = version.split(".")

        if "-" in parts[-1]:
            parts[-1], tag = parts[-1].split("-")

        # 3. Drop leading zeroes from individual version parts
        int_parts = tuple([0 if a == "" else int(a) for a in parts])

        # 4. A maximum of 4 version segments are allowed
        if len(int_parts) > 4:
            raise ValueError(f"Maximum of 4 version segments allowed: "
                             f"'{version}'!")

        # 5. Allow a fourth version segment - "Revision" normally not
        #    allowed in Semantic versions but allowed in NuGet versions.
        #
        #    Version in this case is: <Major>.<Minor>.<Patch>.<Revision>

        # 6. Remove trailing zeros (beyond 3 patch segment)
        #    i.e. If Revision is zero, omit from the normalized string
        if len(int_parts) == 4 and int_parts[3] == 0:
            int_parts = int_parts[0:3]

        # 7. Add missing trailing zeros (below 3 elements).
        #    i.e. 1, 1.0, 1.0.0, and 1.0.0.0 are all accepted and
        #    equal.
        if len(int_parts) < 3:
            int_parts = int_parts + (0,) * (3 - len(int_parts))

        # 8. Reassemble the string for final semantic version validation
        reformed_ver = ".".join((str(num) for num in int_parts))
        if tag is not None:
            reformed_ver += "-" + tag

        # 9. Use semantic_version to further validate the version string
        if len(int_parts) == 4:
            nuget_ver = semantic_version.Version.coerce(reformed_ver)

            # A ValueError will be raised if the version string is invalid
            major, minor, patch, prerelease, build = \
                semantic_version.Version.parse(str(nuget_ver))
        else:
            # A ValueError will be raised if the version string is invalid
            nuget_ver = semantic_version.Version(reformed_ver)
            major, minor, patch, prerelease, build = tuple(nuget_ver)

        logging.info(f"NuGet version parts:\n"
                     f"  Major Version: {major}\n"
                     f"  Minor Version: {minor}\n"
                     f"  Patch Version: {patch}\n"
                     f"  Pre-Release Version {prerelease}\n"
                     f"  Revision (Build) Version: {build}")

        return reformed_ver

    def _fetch_from_nuget_cache(self, package_name: str) -> bool:
        result = False

        #
        # We still need to use Nuget to figure out where the
        # "global-packages" cache is on this machine.
        #
        if self.nuget_cache_path is None:
            cmd = NugetDependency.GetNugetCmd()
            cmd += ["locals", "global-packages", "-list"]
            return_buffer = StringIO()
            if (RunCmd(cmd[0], " ".join(cmd[1:]), outstream=return_buffer) == 0):
                # Seek to the beginning of the output buffer and capture the output.
                return_buffer.seek(0)
                return_string = return_buffer.read()
                self.nuget_cache_path = return_string.strip().replace("global-packages: ", "")

        if self.nuget_cache_path is None:
            logging.info("Nuget was unable to provide global packages cache location.")
            return False
        #
        # If the path couldn't be found, we can't do anything else.
        #
        if not os.path.isdir(self.nuget_cache_path):
            logging.info("Could not determine Nuget global packages cache location.")
            return False

        #
        # Now, try to locate our actual cache path
        nuget_version = self.version
        try:
            nuget_version = NugetDependency.normalize_version(self.version)
        except ValueError:
            logging.error(f"NuGet dependency {self.name} has an invalid "
                          f"version string: {self.version}")

        cache_search_path = os.path.join(
            self.nuget_cache_path, package_name.lower(), nuget_version)
        inner_cache_search_path = os.path.join(cache_search_path, package_name)
        if os.path.isdir(cache_search_path):
            # If we found a cache for this version, let's use it.
            if os.path.isdir(inner_cache_search_path):
                logging.info(self.nuget_cache_path)
                logging.info(
                    "Local Cache found for Nuget package '%s'. Skipping fetch.", package_name)
                shutil.copytree(inner_cache_search_path, self.contents_dir)
                result = True
            # If this cache doesn't match our heuristic, let's warn the user.
            else:
                logging.warning(
                    "Local Cache found for Nuget package '%s', but could not find contents. Malformed?", package_name)

        return result

    def __str__(self) -> str:
        """Return a string representation."""
        return f"NugetDependecy: {self.name}@{self.version}"

    def _attempt_nuget_install(self, install_dir: str, non_interactive: Optional[bool]=True) -> None:
        #
        # fetch the contents of the package.
        #
        package_name = self.name
        cmd = NugetDependency.GetNugetCmd()
        cmd += ["install", self.name]
        cmd += ["-Source", self.source]
        cmd += ["-ExcludeVersion"]
        if non_interactive:
            cmd += ["-NonInteractive"]
        cmd += ["-Version", self.version]
        cmd += ["-Verbosity", "detailed"]
        cmd += ["-OutputDirectory", '"' + install_dir + '"']
        # make sure to capture our output
        output_stream = StringIO()
        ret = RunCmd(cmd[0], " ".join(cmd[1:]), outstream=output_stream)
        output_stream.seek(0)  # return the start of the stream
        # check if we found credential providers
        found_cred_provider = False
        is_unauthorized = False
        for out_line in output_stream:
            line = out_line.strip()
            if line.startswith(("CredentialProvider", "[CredentialProvider")):
                found_cred_provider = True
            if line.endswith("as a credential provider plugin."):
                found_cred_provider = True
            if "401 (Unauthorized)" in line:
                is_unauthorized = True
        # if we fail, then we should retry if we have credential providers
        # we currently steal command input so if we don't have cred providers, we hang
        # this gives cred providers a chance to prompt for input since they don't use stdin
        if ret != 0:
            # If we're in non interactive and we have a credential provider
            if non_interactive and found_cred_provider:  # we should be interactive next time
                self._attempt_nuget_install(install_dir, False)
            else:
                # Only provide this error message if they are not using a credential provider, but receive a 401 error
                if is_unauthorized and not found_cred_provider:
                    logging.warning("[Nuget] A package requires credentials, but you do not have a credential "\
                                    "provider installed.")
                    logging.warning("[Nuget] Please install a credential provider and try again or run the following "\
                                    "command in your terminal to install the package manually:")
                    logging.warning(f"[{' '.join(cmd).replace(' -NonInteractive', '')}]")
                raise RuntimeError(f"[Nuget] We failed to install this version {self.version} of {package_name}")

    def fetch(self) -> None:
        """Fetches the dependency using internal state from the init."""
        package_name = self.name

        # First, check the global cache to see if it's present.
        if super().fetch():
            return

        #
        # Before trying anything with Nuget feeds,
        # check to see whether the package is already in
        # our local cache. If it is, we avoid a lot of
        # time and network cost by copying it directly.
        #
        if self._fetch_from_nuget_cache(package_name):
            self.copy_to_global_cache(self.contents_dir)
            # We successfully found the package in the cache.
            # The published path may change now that the package has been unpacked.
            # Bail.
            self.update_state_file()
            self.published_path = self.compute_published_path()
            return

        #
        # If we are still here, the package wasn't in the cache.
        # We need to ask Nuget to find it.
        #
        temp_directory = self.get_temp_dir()
        self._attempt_nuget_install(temp_directory)

        #
        # Next, copy the contents of the package to the
        # final resting place.
        #
        # Depending on packaging, the package content will be in one of two
        # possible locations:
        # 1. temp_directory\package_name\package_name\
        # 2. temp_directory\package_name\
        #
        source_dir = os.path.join(temp_directory, package_name, package_name)
        if not os.path.isdir(source_dir):
            source_dir = os.path.join(temp_directory, package_name)
        shutil.copytree(source_dir, self.contents_dir)
        self.copy_to_global_cache(self.contents_dir)

        RemoveTree(source_dir)

        #
        # Add a file to track the state of the dependency.
        #
        self.update_state_file()

        #
        # Finally, delete the temp directory.
        #
        RemoveTree(temp_directory)

        # The published path may change now that the package has been unpacked.
        self.published_path = self.compute_published_path()

    def get_temp_dir(self) -> str:
        """Returns the temporary directory the NuGet package is downloaded to."""
        return self.contents_dir + "_temp"

    def clean(self) -> None:
        """Removes the temporary directory the NuGet package is downloaded to."""
        super(NugetDependency, self).clean()
        if os.path.isdir(self.get_temp_dir()):
            RemoveTree(self.get_temp_dir())



# @file usage_report.py
# A report that generates an html report about which repositories INFs (That are consumed for a platform) originate
# from.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A report that generates an html report about which repositories INFs originate from."""
import io
import logging
import pathlib
from argparse import ArgumentParser, Namespace
from typing import Tuple

from edk2toollib.database import Edk2DB, Fv, InstancedInf, Session
from edk2toollib.database import Environment as Env
from sqlalchemy import desc
from sqlalchemy.orm import aliased

from edk2toolext.environment.reporttypes import templates
from edk2toolext.environment.reporttypes.base_report import Report

COLORS = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f','#bcbd22', '#17becf',
          '#aec7e8', '#ffbb78', '#98df8a', '#ff9896', '#c5b0d5',]


class UsageReport(Report):
    """A report that generates a INF usage report for a specific build."""
    def report_info(self) -> Tuple[str, str]:
        """Returns the report standard information.

        Returns:
            (str, str): A tuple of (name, description)
        """
        return ("usage", "Generates a report of INF usage for a specific build. For accurate line counts, run "
                "stuart_parse with the -S flag.")

    def add_cli_options(self, parserobj: ArgumentParser) -> None:
        """Configure command line arguments for this report."""
        parserobj.add_argument("-e", "-env", dest="env_id", action="store",
                               help = "The environment id to generate the report for. Defaults to the latest "
                               "environment.")
        parserobj.add_argument("-o", "-output", dest="output", action="store", default=None,
                               help = "The output file to write the report to. Defaults to 'usage_report.html'.")

    def run_report(self, db: Edk2DB, args: Namespace) -> None:
        """Generate the Usage report."""
        try:
            import plotly.graph_objects as go
            from jinja2 import Environment, FileSystemLoader
        except ImportError as e:
            print(e)
            print("WARNING: This report requires pip modules not installed with edk2-pytool-extensions:")
            print("  Run the following command: `pip install jinja2 plotly`")
            exit(-1)

        # Retrieve needed data from the database and build the data sets.
        with db.session() as session:
            env_id = args.env_id or session.query(Env).order_by(desc(Env.date)).first().id
            version = session.query(Env).filter_by(id=env_id).first().version
            env_vars = self._get_env_vars(session, env_id)
            reports, inf_list = self.generate_data(env_id, session)

        # Build color map for consistent colors across all reports
        color_map = {}
        for idx, key in enumerate({item[0] for item in inf_list}):
            color_map[key] = COLORS[idx % len(COLORS)]

        # Vars for html template
        env = Environment(loader=FileSystemLoader(templates.__path__))
        template = env.get_template("usage_report_template.html")
        report_data = {
            "version": version,
            "env": env_vars,
            "inf_list":  inf_list,
        }

        # Build the pie charts and save them in report_data
        for key, value, title, combine in reports:
            labels = [key for key in value.keys()]
            if combine:
                values = [len(set(value)) for value in value.values()]
            else:
                values = [value[key] for key in value.keys()]
            fig = go.Figure(go.Pie(labels=labels, values=values, hole = .3, title=title, titleposition="top center"))
            fig.update_traces(marker=dict(colors=[color_map[key] for key in value.keys()]))
            # Write the html
            html = io.StringIO()
            fig.write_html(html, full_html=False, include_plotlyjs=False)
            html.seek(0)

            # Add the html to the data dictionary
            report_data[key] = html.read()

        # Open the template and write the html with the report data
        html_output = template.render(**report_data)
        path_out = args.output or report_data["env"].get("PLATFORM_NAME", None) or "usage_report.html"
        if not path_out.endswith(".html"):
            path_out += ".html"

        pathlib.Path(path_out).parent.mkdir(exist_ok=True, parents=True)
        with open(path_out, 'w') as f:
            f.write(html_output)
        logging.info(f"Report written to {path_out}.")

    def generate_data(self, env_id: int, session: Session) -> Tuple[dict, set]:
        """Generates a list of pie chart data.

        Args:
            env_id (int): The environment id the report is generating off of.
            session (Session): The session with a connection to the database.

        Returns:
            (dict, set): (pie chart data, set of all INFs)
        """
        lib_infs = {}
        comp_infs = {}
        total_infs = {}
        lib_lines = {}
        comp_lines = {}
        total_lines = {}

        inf_dict = {}

        # Generate a list of all INFs used by the platform. Do this by grabbing
        # all Components from FVs, then enumerate the INFs that have that
        # component as their component.
        inf_alias = aliased(InstancedInf)
        inf_list = (
            session.query(inf_alias)
                .join(Fv.infs)
                .join(inf_alias, InstancedInf.path == inf_alias.component)
                .filter(Fv.env == env_id)
                .filter(inf_alias.env == env_id)
                .filter(InstancedInf.env == env_id)
                .group_by(inf_alias.path)
                .distinct(inf_alias.path)
                .all()
        )
        final_data = []

        # Finally iterate through each INF and build the datatable we need.
        for inf in inf_list:
            for source in inf.sources:
                if inf.package is None:
                    package_name = "None"
                else:
                    package_name = inf.package.name

                final_data.append((
                    inf.repository.name,
                    package_name,
                    inf.path,
                    source.path,
                    source.code_lines,
                    inf.path == inf.component,
                ))

        for repo, package, inf, _src, line_count, is_component in final_data:
            key = (repo, package, inf)
            current = inf_dict.setdefault(key, (repo, package, inf, 0))
            inf_dict[key] = (repo, package, inf, current[3] + (line_count or 0))

            if is_component:
                inf_d = comp_infs
                src_d = comp_lines
            else:
                inf_d = lib_infs
                src_d = lib_lines

            inf_d.setdefault(repo, []).append(inf)
            src_d[repo] = src_d.get(repo, 0) + (line_count or 0)

        total_infs = self._merge_dicts(lib_infs, comp_infs)
        total_lines = self._merge_dicts(lib_lines, comp_lines)

        # Build the reports
        reports = [
            ("total_pie_chart", total_infs, "Total INF Usage Per Repository", True),
            ("lib_pie_chart", lib_infs, "Library Usage Per Repository", True),
            ("comp_pie_chart", comp_infs, "Component Usage Per Repository", True),
            ("total_src_pie_chart", total_lines, "Total Line Count Per Repository", False),
            ("lib_src_pie_chart", lib_lines, "Library Line Count Per Repository", False),
            ("comp_src_pie_chart", comp_lines, "Component Line Count Per Repository", False),
        ]
        return (reports, set(inf_dict.values()))

    def _get_env_vars(self, session: Session, env_id: int) -> dict[str, str]:
        env_vars = {}
        env = session.query(Env).filter_by(id=env_id).one()
        for value in env.values:
            env_vars[value.key] = value.value
        return env_vars

    def _merge_dicts(self, dict1: dict, dict2: dict) -> dict:
        return_dict = {}
        for key in dict1:
            if key in dict2:
                return_dict[key] = dict1[key] + dict2[key]
            else:
                return_dict[key] = dict1[key]
        for key in dict2:
            if key not in return_dict:
                return_dict[key] = dict2[key]
        return return_dict



# @file edk2_report.py
# An executable that allows a user to select a report and execute it on a given database.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""An executable that allows a user to select a report and execute it on a given database."""
import glob
import logging
import pathlib
import sys
from argparse import ArgumentParser, Namespace
from datetime import datetime

from edk2toollib.database import Edk2DB

from edk2toolext import edk2_logging
from edk2toolext.environment.reporttypes import ComponentDumpReport, CoverageReport, UsageReport

REPORTS = [CoverageReport(), ComponentDumpReport(), UsageReport()]


def setup_logging(verbose: bool) -> None:
    """Setup logging for the tool."""
    logger = logging.getLogger('')
    logger.setLevel(logging.DEBUG if verbose else logging.INFO)
    edk2_logging.setup_section_level()
    edk2_logging.setup_console_logging(logging.DEBUG if verbose else logging.INFO)
    logging.info("Log Started: " + datetime.strftime(datetime.now(), "%A, %B %d, %Y %I:%M%p"))


def parse_args() -> Namespace:
    """Parse the arguments for the tool."""
    parser = ArgumentParser("A tool to generate reports on a edk2 workspace.")
    parser.add_argument('--verbose', '--VERBOSE', '-v', dest="verbose", action='store_true', default=False,
                        help='verbose')
    parser.add_argument('-db', '--database', '--DATABASE', dest='database',
                        default=str(pathlib.Path("Build","DATABASE.db")),
                        help="The database to use when generating reports. Can be a comma separated list of db's to "
                             "merge. Globbing is supported.")

    # Register the report arguments as subparser
    subparsers = parser.add_subparsers(dest='cmd', required=[])
    for report in REPORTS:
        name, description = report.report_info()
        report_parser = subparsers.add_parser(name, description=description)
        report.add_cli_options(report_parser)

    return parser.parse_args()

def main() -> int:
    """Main functionality of the executable."""
    args = parse_args()
    setup_logging(args.verbose)

    logging.warning("stuart_report is in active development. Please report any issues to the edk2-pytool-extensions "
                    "repo.")
    # Verify arguments
    to_merge = []
    database_list = args.database.split(",")
    for database in database_list:
        found = list(glob.glob(database))
        if not found:
            logging.warning(f"No database at path: [{database}]")
        else:
            to_merge.extend(list(glob.glob(database)))

    if not to_merge:
        logging.error("No databases found.")
        return -1

    for database in to_merge:
        if not pathlib.Path(database).exists():
            logging.error(f"Database does not exist: [{database}]")
            return -1

    if len(to_merge) == 1:
        logging.info(f"Single Database file found at: {to_merge[0]}")
        db_path = to_merge[0]
    else:
        logging.error("Multiple Databases not Currently Supported.")
        return -1

    del args.database
    cmd = args.cmd
    del args.cmd

    db = Edk2DB(db_path = db_path)
    for report in REPORTS:
        name, _ = report.report_info()
        if name == cmd:
            return report.run_report(db, args)
    return -1


def go() -> None:
    """Main entry into the report generator tool.

    Sets up the logger for the tool and then runs the tool.
    """
    # setup main console as logger
    logger = logging.getLogger('')
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter("%(levelname)s - %(message)s")
    console = logging.StreamHandler()
    console.setLevel(logging.WARNING)
    console.setFormatter(formatter)
    logger.addHandler(console)

    # call main worker function
    retcode = main()

    if retcode != 0:
        logging.critical("Failed.  Return Code: %d" % retcode)
    else:
        logging.critical("Success!")
    # end logging
    logging.shutdown()
    sys.exit(retcode)


if __name__ == '__main__':
    go()



# @file component_report.py
# A report to print information about a component that could be compiled.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A report to print information about a component that could be compiled."""
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path, PurePath
from typing import Tuple

from edk2toollib.database import Edk2DB, Environment, InstancedInf, Session
from sqlalchemy import desc


class ComponentDumpReport:
    """A report to print information about a component that could be compiled."""
    def report_info(self) -> Tuple[str, str]:
        """Returns the report standard information.

        Returns:
            (str, str): A tuple of (name, description)
        """
        return ("component-libs", "Dumps the library instances used by  component.")

    def add_cli_options(self, parserobj: ArgumentParser) -> None:
        """Configure command line arguments for this report."""
        parserobj.add_argument(dest="component", action="store", help="The component to query.")
        parserobj.add_argument("-o", "--out", dest="file", default=sys.stdout, help="The file, to write the report to."
                               " Defaults to stdout.")
        parserobj.add_argument("-d", "--depth", dest="depth", type=int, default=999, help="The depth to recurse when "
                               "printing libraries used.")
        parserobj.add_argument("-f", "--flatten", dest="flatten", action="store_true",
                               help="Flatten the list of libraries used in the component.")
        parserobj.add_argument("-s", "--sort", dest="sort", action="store_true",
                               help="Sort the libraries listed in alphabetical order.")
        parserobj.add_argument("-e", "--env", dest="env_id", action="store",
                               help="The environment id to generate the report for.")

    def run_report(self, db: Edk2DB, args: Namespace) -> None:
        """Runs the report."""
        if isinstance(args.file, str):
            if Path(args.file).exists():
                Path(args.file).unlink()
            self.file = open(args.file, 'w+')
        else:
            self.file = args.file

        self.depth = args.depth
        self.sort = args.sort

        self.component = PurePath(args.component).as_posix()

        with db.session() as session:
            self.env_id = args.env_id or session.query(Environment).order_by(desc(Environment.date)).first().id
            component = (
                session
                    .query(InstancedInf)
                    .filter_by(env=self.env_id, cls=None)
                    .filter(InstancedInf.path.like(f'%{self.component}%'))
                    .one()
            )

            if args.flatten:
                return self.print_libraries_flat(component.path, session)

            libraries = component.libraries
            if self.sort:
                libraries = sorted(libraries, key=lambda x: x.cls)

            print(component.path, file=self.file)
            for library in libraries:
                self.print_libraries_recursive(library, [], session)

    def print_libraries_recursive(self, library: InstancedInf, visited: list, session: Session, depth: int = 0) -> None:
        """Prints the libraries used in a provided library / component."""
        if depth >= self.depth:
            return

        library_class = library.cls
        library_instance = library.path

        print(f'{"  "*depth}- {library_class}| {library_instance or "NOT FOUND IN DSC"}', file=self.file)

        if library_instance is None:
            return
        libraries = library.libraries

        if self.sort:
            libraries = sorted(libraries, key=lambda x: x[1])

        for library in libraries:
            if library in visited:
                continue
            visited.append(library)
            self.print_libraries_recursive( library, visited.copy(), session, depth=depth+1)
        return

    def print_libraries_flat(self, component: str, session: Session) -> None:
        """Prints the libraries used in a provided component."""
        libraries = (
            session
                .query(InstancedInf)
                .filter_by(env = self.env_id, component = component)
                .filter(InstancedInf.cls.isnot(None))
                .all()
        )

        length = max(len(library.cls) for library in libraries)
        if self.sort:
            libraries = sorted(libraries, key=lambda x: x.cls)

        for library in libraries:
            print(f'- {library.cls:{length}}| {library.path}', file=self.file)



# @file external_dependencies.py
# This module contains helper objects that can manipulate,
# retrieve, validate, and clean external dependencies for the
# build environment.
#
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""This module contains helper object for manipulating external dependencies.

These manipulations include retrieving, validating, and cleaning external
dependencies for the build environment.
"""

import hashlib
import logging
import os
import shutil
from typing import Optional

import yaml
from edk2toollib.utility_functions import GetHostInfo, RemoveTree

from edk2toolext.environment import version_aggregator


class ExternalDependency(object):
    """Baseclass to allow downloading external dependencies during the update phase.

    Specific External dependency types (git, nuget, etc.) are subclassed from
    this class. Additional external dependency types may be created.

    Attributes:
        scope (str): Determines if descriptor is included on a particular build.
        type (str): Type of ext_dep.
        name (str): Name of ext_dep, used to name the folder the ext_dep will be unpacked in to
        source (str): Source to query for ext_dep.
        version (str): Version string to keep track of what version is currently installed.
        flags (list[str]): Flags dictating what actions should be taken once this dependency is resolved
                           More info: (docs/feature_extdep/)
        var_name (str): Used with set_*_var flag. Determines name of var to be set.

    !!! tip
        The attributes are what must be described in the ext_dep yaml file!
    """

    def __init__(self: str, descriptor: dict) -> None:
        """Inits a web dependency based off the provided descriptor."""
        super(ExternalDependency, self).__init__()

        #
        # Set the data for this object.
        #
        self.scope = descriptor['scope']
        self.type = descriptor['type']
        self.name = descriptor['name']
        self.source = descriptor['source']
        self.version = descriptor['version']
        self.flags = descriptor.get('flags', None)
        self.var_name = descriptor.get('var_name', None)
        self.error_msg = descriptor.get('error_msg', None)
        self.global_cache_path = None

        self.descriptor_location = os.path.dirname(
            descriptor['descriptor_file'])
        self.contents_dir = os.path.join(
            self.descriptor_location, self.name + "_extdep")
        self.state_file_path = os.path.join(
            self.contents_dir, "extdep_state.yaml")
        self.published_path = self.compute_published_path()

    def set_global_cache_path(self, global_cache_path: str) -> 'ExternalDependency':
        """Sets the global cache path to locate already downloaded dependencies.

        Arguments:
            global_cache_path (str): directory of the global cache
        """
        self.global_cache_path = os.path.abspath(global_cache_path)
        return self

    def compute_published_path(self) -> str:
        """Determines the published path."""
        new_published_path = self.contents_dir

        if self.flags and "host_specific" in self.flags and self.verify():
            host = GetHostInfo()

            logging.info("Computing path for {0} located at {1} on {2}".format(self.name,
                                                                               self.contents_dir,
                                                                               str(host)))

            acceptable_names = []

            # we want to list all the possible folders we would be comfortable using
            # and then check if they are present.
            # The "ideal" directory name is OS-ARCH-BIT
            acceptable_names.append("-".join((host.os, host.arch, host.bit)))
            acceptable_names.append("-".join((host.os, host.arch)))
            acceptable_names.append("-".join((host.os, host.bit)))
            acceptable_names.append("-".join((host.arch, host.bit)))
            acceptable_names.append(host.os)
            acceptable_names.append(host.arch)
            acceptable_names.append(host.bit)

            new_published_path = None
            for name in acceptable_names:
                dirname = os.path.join(self.contents_dir, name)
                if os.path.isdir(dirname):
                    logging.info("{0} was found!".format(dirname))
                    new_published_path = dirname
                    break
                logging.debug("{0} does not exist".format(dirname))

            if new_published_path is None:
                logging.error(f"{self.name} is host specific, but does not appear to have support for {str(host)}.")
                logging.error(f"Verify support for detected host: {str(host)} and contact dependency provider to add "\
                              "support.")
                logging.error("Otherwise, delete the external dependency directory to reset.")

                new_published_path = self.contents_dir

        if self.flags and "include_separator" in self.flags:
            new_published_path += os.path.sep

        return new_published_path

    def clean(self: str) -> None:
        """Removes the local directory for the external dependency."""
        logging.debug("Cleaning dependency directory for '%s'..." % self.name)
        if os.path.isdir(self.contents_dir):
            RemoveTree(self.contents_dir)

    def determine_cache_path(self) -> Optional[str]:
        """Determines the cache path is global_cache_path is not none."""
        result = None
        if self.global_cache_path is not None and os.path.isdir(self.global_cache_path):
            subpath_calc = hashlib.sha1()
            subpath_calc.update(self.version.encode('utf-8'))
            subpath_calc.update(self.source.encode('utf-8'))
            subpath = subpath_calc.hexdigest()
            result = os.path.join(self.global_cache_path, self.type, self.name, subpath)
        return result

    def fetch(self) -> bool:
        """Fetches the dependency using internal state from the init."""
        cache_path = self.determine_cache_path()
        if cache_path is None or not os.path.isdir(cache_path):
            return False
        logging.debug("Found %s extdep '%s' in global cache." % (self.type, self.name))
        self.copy_from_global_cache(self.contents_dir)
        self.published_path = self.compute_published_path()
        self.update_state_file()
        return True

    def copy_from_global_cache(self, dest_path: str) -> None:
        """Copies the dependency from global cache if present.

        Arguments:
            dest_path (str): path to copy to
        """
        cache_path = self.determine_cache_path()
        if cache_path is None:
            return
        if os.path.isdir(cache_path):
            shutil.copytree(cache_path, dest_path, dirs_exist_ok=True)

    def copy_to_global_cache(self, source_path: str) -> None:
        """Copies the dependency to global cache if present.

        Arguments:
            source_path (str): source to copy into global cache.
        """
        cache_path = self.determine_cache_path()
        if cache_path is None:
            return
        if not os.path.isdir(cache_path):
            os.makedirs(cache_path)
        else:
            shutil.rmtree(cache_path)
        shutil.copytree(source_path, cache_path, dirs_exist_ok=True)

    def verify(self) -> int:
        """Verifies the dependency was successfully downloaded."""
        result = True
        state_data = None

        # See whether or not the state file exists.
        if not os.path.isfile(self.state_file_path):
            result = False

        # Attempt to load the state file.
        if result:
            with open(self.state_file_path, 'r') as file:
                try:
                    state_data = yaml.safe_load(file)
                except Exception:
                    pass
        if state_data is None:
            result = False

        # If loaded, check the version.
        if result and state_data['version'] != self.version:
            result = False

        logging.debug("Verify '%s' returning '%s'." % (self.name, result))
        return result

    def report_version(self) -> None:
        """Reports the version of the external dependency."""
        version_aggregator.GetVersionAggregator().ReportVersion(self.name,
                                                                self.version,
                                                                version_aggregator.VersionTypes.INFO,
                                                                self.descriptor_location)

    def update_state_file(self) -> None:
        """Updates the file representing the state of the dependency."""
        with open(self.state_file_path, 'w+') as file:
            yaml.dump({'version': self.version}, file)


def ExtDepFactory(descriptor: dict) -> 'ExternalDependency':
    """External Dependency Factory capable of generating each type of dependency.

    !!! Note
        Ensure all external dependencies are imported in this class to avoid errors.
    """
    from edk2toolext.environment.extdeptypes.az_cli_universal_dependency import AzureCliUniversalDependency
    from edk2toolext.environment.extdeptypes.git_dependency import GitDependency
    from edk2toolext.environment.extdeptypes.nuget_dependency import NugetDependency
    from edk2toolext.environment.extdeptypes.web_dependency import WebDependency
    if descriptor['type'] == NugetDependency.TypeString:
        return NugetDependency(descriptor)
    elif descriptor['type'] == WebDependency.TypeString:
        return WebDependency(descriptor)
    elif descriptor['type'] == GitDependency.TypeString:
        return GitDependency(descriptor)
    elif descriptor['type'] == AzureCliUniversalDependency.TypeString:
        AzureCliUniversalDependency.VerifyToolDependencies()
        return AzureCliUniversalDependency(descriptor)

    raise ValueError("Unknown extdep type '%s' requested!" % descriptor['type'])



# @file base_report.py
# An interface to create custom reports with.
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""An interface to create custom reports with."""
from argparse import ArgumentParser, Namespace
from typing import Tuple

from edk2toollib.database import Edk2DB


class Report:
    """The interface to create custom reports."""
    def report_info(self) -> Tuple[str, str]:
        """Returns the report standard information.

        Returns:
            (str, str): A tuple of (name, description)
        """
        raise NotImplementedError

    def add_cli_options(self, parserobj: ArgumentParser) -> None:
        """Configure command line arguments for this report."""
        return

    def run_report(self, db: Edk2DB, args: Namespace) -> None:
        """Generate a report."""
        raise NotImplementedError



# @file plugin_manager.py
# This module contains code that supports Build Plugins
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""This module contains code that supports Build Plugins."""

import importlib
import logging
import os
import sys
import warnings

from edk2toolext.environment import shell_environment


class PluginDescriptor(object):
    """Plugin Descripter.

    Attributes:
        descripter(Dict): descriptor
        Obj (obj): Object
        Name (str): name attribute from descriptor
        Module (obj): module attribute from descriptor
    """
    def __init__(self, t:dict) -> None:
        """Inits the Plugin descriptor with the Descriptor."""
        self.descriptor = t
        self.Obj = None
        self.Name = t["name"]
        self.Module = t["module"]

    def __str__(self) -> str:
        """String representation of the plugin descriptor."""
        return "PLUGIN DESCRIPTOR:{0}".format(self.Name)


class PluginManager(object):
    """A class that manages all plugins in the environment.

    Attributes:
        Descriptors (List[PluginDescriptor]): list of plugin descriptors
    """
    def __init__(self) -> None:
        """Inits an empty plugin manager."""
        self.Descriptors = []

    def SetListOfEnvironmentDescriptors(self, newlist: list) -> int:
        """Passes a tuple of environment descriptor dictionaries to be loaded as plugins."""
        env = shell_environment.GetBuildVars()
        failed = []
        if newlist is None:
            return []
        for a in newlist:
            b = PluginDescriptor(a)
            if (self._load(b) == 0):
                val = env.GetValue(b.Module.upper())
                if val and val == "skip":
                    logging.info(f"{b.Module} turned off by environment variable")
                    continue
                self.Descriptors.append(b)
            else:
                failed.append(a)
        return failed

    def GetPluginsOfClass(self, classobj: type) -> list[object]:
        """Return list of all plugins of a given class.

        Returns:
            list[object]: plugin instances of the requested class
        """
        temp = []
        for a in self.Descriptors:
            if (isinstance(a.Obj, classobj)):
                temp.append(a)
        return temp

    def GetAllPlugins(self) -> list[object]:
        """Return list of all plugins."""
        return self.Descriptors

    def _load(self, PluginDescriptor: 'PluginDescriptor') -> int:
        """Load and instantiate the plugin.

        Args:
            PluginDescriptor(PluginDescriptor): the plugin descriptor
        """
        PluginDescriptor.Obj = None

        py_file_path = PluginDescriptor.descriptor["module"] + ".py"
        py_module_path = os.path.join(os.path.dirname(os.path.abspath(
            PluginDescriptor.descriptor["descriptor_file"])), py_file_path)
        py_module_name = "UefiBuild_Plugin_" + PluginDescriptor.descriptor["module"]

        logging.debug("Loading Plugin from %s", py_module_path)

        try:
            spec = importlib.util.spec_from_file_location(
                py_module_name, py_module_path)
            module = importlib.util.module_from_spec(spec)
            sys.modules[py_module_name] = module

            py_module_dir = os.path.dirname(py_module_path)
            if py_module_dir not in sys.path:
                sys.path.append(py_module_dir)

            # Turn on Deprecation warnings for code in the plugin
            warnings.filterwarnings("default", category=DeprecationWarning, module=module.__name__)

            spec.loader.exec_module(module)
        except Exception:
            exc_info = sys.exc_info()
            logging.error("Failed to import plugin: %s",
                          py_module_path, exc_info=exc_info)
            return -1

        # Instantiate the plugin
        try:
            obj = getattr(module, PluginDescriptor.descriptor["module"])
            PluginDescriptor.Obj = obj()
        except AttributeError:
            exc_info = sys.exc_info()
            logging.error("Failed to instantiate plugin: %s",
                          py_module_path, exc_info=exc_info)
            return -1

        return 0



##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A collection of reports that can be generated from a provided database."""

from edk2toolext.environment.reporttypes.component_report import ComponentDumpReport  # noqa: F401
from edk2toolext.environment.reporttypes.coverage_report import CoverageReport  # noqa: F401
from edk2toolext.environment.reporttypes.usage_report import UsageReport  # noqa: F401
